program logdecon
  use sep
  use FT
  use hybrid
  use axisIO
  use quantile_mod

  implicit none

  interface 
    SUBROUTINE calc_new_norm(inc_alpha,r,dr,wrk_2d_2,wrk_2d_3,output,threshold_traces,weight)

      implicit none
      optional :: weight
      double complex,intent(in)	:: inc_alpha
      double complex,intent(out)	:: output
      complex,dimension(:,:) 	:: r,dr
      complex,dimension(:,:) 	:: wrk_2d_2,wrk_2d_3
      real,dimension(:,:),intent(in) :: weight
      real ,dimension(:),intent(in) ::threshold_traces
    end SUBROUTINE 

    SUBROUTINE calc_current_norm(r,wrk_2d_2,wrk_2d_3,output,threshold_traces,weight)
      use hybrid
      implicit none
      optional :: weight
      double complex,intent(out)	:: output
      complex,dimension(:,:) 	:: r
      complex,dimension(:,:) 	:: wrk_2d_2,wrk_2d_3
      real ,dimension(:,:),intent(in) :: weight
      real ,dimension(:),intent(in) ::threshold_traces
    end SUBROUTINE

  end interface

  complex, allocatable :: dd(:,:)
  double complex, allocatable :: uu(:),duu(:),euu(:)
  complex, allocatable::d(:,:),r(:,:),dr(:,:),drr(:,:),prv_r(:,:),wr(:,:)
  double complex, allocatable::u(:),du(:)
  complex, allocatable::rr(:,:),wrk_2d_2(:,:),wrk_2d_3(:,:),wrk_2d_4(:,:)
  complex, allocatable::wrk2(:),wrk3(:),wrk4(:)
  integer :: nthreads,ithreads
  double precision :: f_d_scale,b_d_scale
  double complex::temp1,temp2,alpha,inc_alpha,prv_r_norm,temp3,ratio
  double complex::temp4(1),temp5
  logical ::goto_updata_dr
  integer :: nt,nx,nx_org
  integer :: it,ix
  integer :: iter,niter
  integer :: iter_alpha,niter_alpha
  real    :: threshold, quantile_value
  real, allocatable   :: data(:,:),res(:,:)
  real, allocatable   :: res_norm(:)

  integer :: npad
  integer :: n_sym_init
  integer :: weight_n1,weight_n2,weight_n3,weight_ix
  logical :: is_weight
  real, allocatable :: weight(:,:) ,mask_du(:),whwr(:)

  double precision:: input_file_scale

  real :: real_input_file_scale

  character(len=256) :: csz_threshold, csz_niter, csz_niter_alpha, csz_u_file, csz_pad, csz_res_norm_file, csz_filter_file,csz_weight_file,csz_input_file_scale,csz_l1_l2_file, csz_quantile, csz_wr_file, csz_du_file, csz_linear_r_file, csz_mask_du_file, csz_index_file, csz_whwr_file, csz_sym_init, csz_sym_init_r_file, csz_sym_near_du
  type(myaxis)       :: axis_1,axis_2,axis_3,axis_4
  character(len=256) :: u_file,res_norm_file, filter_file, weight_file, l1_l2_file, wr_file, du_file, linear_r_file, mask_du_file, index_file, whwr_file, sym_init_r_file
  integer n_sym_near_du

  integer, external	:: omp_get_thread_num,omp_get_num_threads
  logical :: input_n3
  real, allocatable :: threshold_traces(:)
  real :: convergence_index(1)

  double complex :: dc_temp

!   double complex :: dc_one=1.

  csz_threshold = "threshold"
  csz_niter="niter"
  csz_niter_alpha="niter_alpha"
  csz_u_file="u_file"
  csz_du_file="du_file"
  csz_mask_du_file="mask_du_file"
  csz_linear_r_file="linear_r_file"
  csz_index_file="index_file"
  csz_pad="pad"
  csz_res_norm_file="res_norm_file"
  csz_filter_file="filter_file"

  csz_weight_file="weight_file"
  csz_input_file_scale="input_file_scale"
  csz_l1_l2_file="l1_l2_file"
  csz_quantile="quantile"
  csz_wr_file="wr_file"
  csz_whwr_file="whwr_file"

  csz_sym_init="sym_init"
  csz_sym_init_r_file="sym_init_r_file"
  csz_sym_near_du="sym_near_du"

  call sep_init()

  call from_param (csz_threshold, threshold, 0.1)
  call from_param (csz_niter, niter, 10)
  call from_param (csz_niter_alpha, niter_alpha, 5)
  call from_param (csz_u_file, u_file, 'NULL')
  call from_param (csz_du_file, du_file, 'NULL')
  call from_param (csz_mask_du_file, mask_du_file, 'NULL')
  call from_param (csz_linear_r_file, linear_r_file, 'NULL')
  call from_param (csz_index_file, index_file, 'NULL')
  call from_param (csz_pad, npad, 1)
  call from_param (csz_res_norm_file, res_norm_file, 'NULL')
  call from_param (csz_filter_file, filter_file, 'NULL')
  call from_param (csz_weight_file, weight_file, 'NULL')
  call from_param (csz_input_file_scale,real_input_file_scale,0.)
  call from_param (csz_l1_l2_file, l1_l2_file, 'NULL')
  call from_param (csz_quantile,quantile_value,0.)
  call from_param (csz_wr_file, wr_file, 'NULL')
  call from_param (csz_whwr_file, whwr_file, 'NULL')
  call from_param (csz_sym_init, n_sym_init, 0)
  call from_param (csz_sym_init_r_file, sym_init_r_file, 'NULL')
  call from_param (csz_sym_near_du, n_sym_near_du, 0)



  if (quantile_value.lt.0. .or. quantile_value .gt.1.) then
    write(0,*) "(quantile_value.lt.0. .or. quantile_value .gt.1.)"
    quantile_value=0.
    call to_history (csz_quantile, quantile_value)
  end if
  if (quantile_value.gt.0.) then
    write(0,*) "quantile_value=",quantile_value
    threshold=1.
    write(0,*) "The threshold has been set to 1.0"
    call to_history (csz_threshold, threshold)
  end if


  if (npad.lt.1) npad=1

  call hinaxis(axis_1,1)
  call hinaxis(axis_2,2)
  axis_1%n=axis_1%n*npad
  nt = axis_1%n  
  nx = axis_2%n

  input_n3=.false.
  nx_org=nx
  call hinaxis(axis_3,3)
!   call assert(axis_3%n.eq.1 ,'axis_3%n.eq.1')
  if (axis_3%n .gt. 1) then
    nx=nx*axis_3%n
    input_n3=.true.
    call hinaxis(axis_4,4)
    call assert(axis_4%n.eq.1 ,'axis_4%n.eq.1')
  end if

  call houtaxis(axis_1,1)

  call houtaxis(axis_2,2)


  call assert(n_sym_near_du.lt.nt/2 .and. n_sym_near_du.ge.0,'n_sym_near_du.lt.nt/2 .and. n_sym_near_du.ge.0')


  if (input_n3.eq..false.) then
    axis_3%n=niter+1
    axis_3%o=0.
    axis_3%d=1.
    axis_3%l='niter'
    call houtaxis(axis_3,3)
  else
    axis_4%n=niter+1
    axis_4%o=0.
    axis_4%d=1.
    axis_4%l='niter'
    call houtaxis(axis_4,4)
  end if




  if (u_file.ne.'NULL') then
      axis_2%n=1
      axis_3%n=niter
      axis_3%o=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,u_file)
    call houtaxis(axis_2,2,u_file)
    call houtaxis(axis_3,3,u_file)
  end if

  if (du_file.ne.'NULL') then
      axis_2%n=1
      axis_3%n=niter
      axis_3%o=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,du_file)
    call houtaxis(axis_2,2,du_file)
    call houtaxis(axis_3,3,du_file)
  end if

  if (linear_r_file.ne.'NULL') then
      axis_2%n=nx
      axis_3%n=niter
      axis_3%o=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,linear_r_file)
    call houtaxis(axis_2,2,linear_r_file)
    call houtaxis(axis_3,3,linear_r_file)
  end if

  if (filter_file.ne.'NULL') then
      axis_2%n=1
      axis_2%o=1.
      axis_2%l='none'
      axis_3%n=niter
      axis_3%o=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,filter_file)
    call houtaxis(axis_2,2,filter_file)
    call houtaxis(axis_3,3,filter_file)
  end if

  if (res_norm_file.ne.'NULL') then
      axis_3%n=niter
      axis_3%o=1.
      axis_3%l='iteration'
    call houtaxis(axis_3,1,res_norm_file)
  end if

  if (wr_file.ne.'NULL') then
  call hinaxis(axis_1,1)
  call hinaxis(axis_2,2)
      axis_2%n=nx
      axis_3%n=niter
      axis_3%o=1.
      axis_3%d=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,wr_file)
    call houtaxis(axis_2,2,wr_file)
    call houtaxis(axis_3,3,wr_file)
  end if

  if (whwr_file.ne.'NULL') then
  call hinaxis(axis_1,1)
  call hinaxis(axis_2,2)
      axis_2%n=1
      axis_3%n=niter
      axis_3%o=1.
      axis_3%d=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,whwr_file)
    call houtaxis(axis_2,2,whwr_file)
    call houtaxis(axis_3,3,whwr_file)
  end if



  if (sym_init_r_file.ne.'NULL') then
  call hinaxis(axis_1,1)
  call hinaxis(axis_2,2)
      axis_2%n=nx
      axis_3%n=1
      axis_3%o=1.
      axis_3%d=1.
      axis_3%l='iteration'
    call houtaxis(axis_1,1,sym_init_r_file)
    call houtaxis(axis_2,2,sym_init_r_file)
    call houtaxis(axis_3,3,sym_init_r_file)
  end if

  if (index_file.ne.'NULL') then
      axis_1%n=niter
      axis_1%o=1.
      axis_1%d=1.
      axis_1%l='iteration'
      axis_2%n=1
      axis_3%n=1
    call houtaxis(axis_1,1,index_file)
    call houtaxis(axis_2,2,index_file)
    call houtaxis(axis_3,3,index_file)
  end if

  if (l1_l2_file.ne.'NULL') then
      axis_1%n=niter
      axis_1%o=1.
      axis_1%d=1.
      axis_1%l='iteration'
      axis_2%n=1
      axis_3%n=1
    call houtaxis(axis_1,1,l1_l2_file)
    call houtaxis(axis_2,2,l1_l2_file)
    call houtaxis(axis_3,3,l1_l2_file)
  end if


  allocate(mask_du(nt))

  if (mask_du_file.ne.'NULL') then
    call hinaxis(axis_1,1,mask_du_file)
    call hinaxis(axis_2,2,mask_du_file)
    call hinaxis(axis_3,3,mask_du_file)
    weight_n1=axis_1%n
    weight_n2=axis_2%n
    weight_n3=axis_3%n

    

    call assert(axis_1%n.eq.nt,'mask_du_file n1 .eq.nt')
    call assert(axis_2%n.eq.1 .and. axis_3%n.eq.1,'mask_du_file n2 n3 .eq.1')

    call sep_read(mask_du,mask_du_file)
  else
    mask_du=1.
  end if

  if (weight_file.ne.'NULL') then
    call hinaxis(axis_1,1,weight_file)
    call hinaxis(axis_2,2,weight_file)
    call hinaxis(axis_3,3,weight_file)
    weight_n1=axis_1%n
    weight_n2=axis_2%n
    weight_n3=axis_3%n

    if (axis_3%n .gt. 1) then
      weight_n2=weight_n2*axis_3%n
    end if

    call assert(weight_n1.eq.nt,'weight_n1.eq.nt')
    call assert(weight_n2.eq.nx .or. weight_n2.eq.nx_org,'weight_n2.eq.nx .or. weight_n2.eq.nx_org')
    write(0,*) 'debug weight_n2,nx_org,nx',weight_n2,nx_org,nx
!     call assert(weight_n3.eq.1 ,'weight_n3.eq.1')

    allocate(weight(weight_n1,weight_n2))
    is_weight=.true.
    call sep_read(weight,weight_file)
  else
    is_weight=.false.
  end if





  f_d_scale=1.
  b_d_scale=1./nt

   !$OMP PARALLEL
!   call omp_get_dynamic()
!   call OMP_SET_NUM_THREADS(8)
  nthreads = omp_get_num_threads()
!   nthreads = 8
  !$OMP END PARALLEL
  write(0,*) "*******************"
  write(0,*) "* NUM OF THREADS = ",nthreads
  write(0,*) "*******************"

  allocate(dd(nt,nx),uu(nt),duu(nt),drr(nt,nx),d(nt,nx),r(nt,nx),dr(nt,nx),u(nt),du(nt),rr(nt,nx),data(nt,nx),res(nt,nx),wrk2(nt),wrk3(nt),wrk4(nt),res_norm(niter),euu(nt),prv_r(nt,nx),wrk_2d_2(nt,nthreads),wrk_2d_3(nt,nthreads),wrk_2d_4(nt,nthreads),wr(nt,nx),threshold_traces(nx),whwr(nt))



  call sep_read(data)

  if (real_input_file_scale.eq.0.) then
    input_file_scale=maxval(abs(data))/3
  else
    input_file_scale=real_input_file_scale
  end if 

  write(0,*) 'maxval(abs(data))',maxval(abs(data)),'input_file_scale',input_file_scale
!   data=data/maxval(abs(data))*3
  data=data/input_file_scale

  !threshold=threshold/input_file_scale
!   call hybrid_init(threshold)

  d(:,:)=data(:,:)

  call sep_write(real(d))

!!!example     # $OMP parallel do private(inode,iwa,iz) ordered schedule(dynamic,1) num_threads(nodes)

  !$OMP parallel do private(ix) 
  do ix=1,nx
    call FT1D(d(:,ix),dd(:,ix),nt,1,f_d_scale)
  end do
  !$OMP END PARALLEL DO

  if (n_sym_init.eq. 0) then
    u=0.
    u(1)=0. !2*log(1.0)
    euu=0.
    rr(:,:)=dd(:,:) !rr=rr=euu*dd=1*dd=dd
  else
    uu=0.
        !$OMP parallel do private(it,ix) 
        do it=1,nt
          do ix=1,nx
            euu(it)=euu(it)+cabs(dd(it,ix))*cabs(dd(it,ix))
          end do
        end do
        !$OMP END PARALLEL DO
        euu(:)=euu(:)/nx
        euu(:)=1.0/sqrt(euu(:))
!     u(1)=0. !2*log(1.0)
    uu=-log(euu)
    !$OMP parallel do private(ix) 
    do ix=1,nx
      rr(:,ix)=euu(:)*dd(:,ix) !rr=rr=euu*dd
    end do
    !$OMP END PARALLEL DO

    if (sym_init_r_file.ne.'NULL') then
      !$OMP parallel do private(ix) 
      do ix=1,nx
        call FT1D(rr(:,ix),r(:,ix),nt,-1,b_d_scale)
      end do
      !$OMP END PARALLEL DO
      call sep_write(real(real(r)),sym_init_r_file)
    end if

  end if

  
!   r(:,:)=d(:,:)
  prv_r=0.
  prv_r_norm=0.
  temp3=0.
  ratio=0.

  do iter=1,niter
    write(0,*) 'Begin iter',iter
!     write(0,*) u
!     call D_FT1D(u(:),uu(:),nt,1,f_d_scale) !do not need because it will be done in previous loop
    du=0.
    duu=0.
    whwr=0.

!     write(0,*) 'max imag(u)=',maxval(abs(imag(u)))
!     write(0,*) 'max real(u)=',maxval(abs(real(u)))

!quantile
    if (is_weight .eq. .true. ) then
      if ((quantile_value .ne. 0.)) then
        write (0,*) "quantile"
        call assert(threshold .eq. 1. ,'threshold .eq. 1. ')
!         if (iter) then
        !$OMP parallel do private(ix,ithreads,weight_ix) 
        do ix=1,nx
          weight_ix=mod(ix-1,weight_n2)+1
          ithreads=omp_get_thread_num()+1
          call FT1D(rr(:,ix),r(:,ix),nt,-1,b_d_scale)
          wrk_2d_2(:,ithreads)=r(:,ix)*weight(:,weight_ix)
!           write (0,*) "quantile",ix
          
          threshold_traces(ix)=quantile(nint(quantile_value*nt),abs(wrk_2d_2(:,ithreads)))
!           write (0,*) "quantile",ix,,threshold_traces(ix),nint(quantile_value*nt)
          if (threshold_traces(ix).eq. 0.) then
            threshold_traces(ix)=threshold
          end if
          
        end do
        !$OMP END PARALLEL DO
        write (0,*) "quantile done"
        write (0,*) "max ,threshold_traces = ",maxval(threshold_traces)
        write (0,*) "min ,threshold_traces = ",minval(threshold_traces)
!         write (0,*) "max "
        !!!! important
        !!!!!
        !!!!!!
!         weight_scale_factor=weight_scale_factor/maxval(weight_scale_factor)
!         call hybrid_init(threshold/maxval(weight_scale_factor))
      else
          threshold_traces(:)=threshold
      end if
    else
      if ((quantile_value .ne. 0.)) then
        write (0,*) "quantile"
        call assert(threshold .eq. 1. ,'threshold .eq. 1. ')
!         if (iter) then
        !$OMP parallel do private(ix,ithreads) 
        do ix=1,nx
          ithreads=omp_get_thread_num()+1
          call FT1D(rr(:,ix),r(:,ix),nt,-1,b_d_scale)
          wrk_2d_2(:,ithreads)=r(:,ix)
!           write (0,*) "quantile",ix
          
          threshold_traces(ix)=quantile(nint(quantile_value*nt),abs(wrk_2d_2(:,ithreads)))
!           write (0,*) "quantile",ix,,threshold_traces(ix),nint(quantile_value*nt)
          if (threshold_traces(ix).eq. 0.) then
            threshold_traces(ix)=threshold
          end if
          
        end do
        !$OMP END PARALLEL DO
        write (0,*) "quantile done"
        write (0,*) "max ,threshold_traces = ",maxval(threshold_traces)
        write (0,*) "min ,threshold_traces = ",minval(threshold_traces)
      else
          threshold_traces(:)=threshold
      end if
    end if




    if (is_weight.eq..false.) then
      !$OMP parallel do private(ix,ithreads) REDUCTION(+:duu)
      do ix=1,nx
        ithreads=omp_get_thread_num()+1
!       do not need to do:  rr(:,ix)=dd(:,ix)*euu(:) !rr=euu*dd 
        call FT1D(rr(:,ix),r(:,ix),nt,-1,b_d_scale)
        r(:,ix)=real(r(:,ix)) ! todo: do we need to get real of r? yes it is inportant!!!
        call hybrid_d1(r(:,ix),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H'(r)
        call FT1D(wrk_2d_2(:,ithreads),wrk_2d_3(:,ithreads),nt,1,f_d_scale) !wrk3=FT(H'(r))
        whwr=whwr+wrk_2d_2(:,ithreads)
        call FT1D(r(:,ix),wrk_2d_4(:,ithreads),nt,1,f_d_scale) !wrk4=FT(r)
        duu=duu+conjg(wrk_2d_4(:,ithreads))*wrk_2d_3(:,ithreads) !duu=corr(H'(r),r)
      end do
      !$OMP END PARALLEL DO
    else
      !$OMP parallel do private(ix,ithreads,weight_ix) REDUCTION(+:duu)
      do ix=1,nx
        weight_ix=mod(ix-1,weight_n2)+1
        ithreads=omp_get_thread_num()+1
!       do not need to do:  rr(:,ix)=dd(:,ix)*euu(:) !rr=euu*dd 
        call FT1D(rr(:,ix),r(:,ix),nt,-1,b_d_scale)
        r(:,ix)=real(r(:,ix)) ! todo: do we need to get real of r? yes it is inportant!!!
        wr(:,ix)=r(:,ix)*weight(:,weight_ix)
        call hybrid_d1(wr(:,ix),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H'(w*r)
        wrk_2d_2(:,ithreads)=wrk_2d_2(:,ithreads)*weight(:,weight_ix) !wrk2=w*H'(w*r)
        whwr=whwr+wrk_2d_2(:,ithreads)
        call FT1D(wrk_2d_2(:,ithreads),wrk_2d_3(:,ithreads),nt,1,f_d_scale) !wrk3=FT(w*H'(w*r))
        call FT1D(r(:,ix),wrk_2d_4(:,ithreads),nt,1,f_d_scale) !wrk4=FT(r)
!         call FT1D(conjg(wrk_2d_4(:,ithreads))*wrk_2d_3(:,ithreads),wrk_2d_2(:,ithreads),nt,-1,b_d_scale) !wrk4=delta_u
!         wrk_2d_3(:,ithreads)=wrk_2d_2(:,ithreads) !*weight(:,weight_ix)
!         call FT1D(wrk_2d_3(:,ithreads),wrk_2d_2(:,ithreads),nt,1,f_d_scale) !wrk4=FT(r)
!         duu=duu+wrk_2d_2(:,ithreads)
         duu=duu+conjg(wrk_2d_4(:,ithreads))*wrk_2d_3(:,ithreads) !duu=corr(r,w*H'(w*r))
      end do
      !$OMP END PARALLEL DO
    end if


    call sep_write(real(real(whwr)),whwr_file)

    if(wr_file.ne.'NULL' ) then
      call sep_write(real(real(wr)),wr_file)
    end if

    duu=duu/nx

    call D_FT1D(duu(:),du(:),nt,-1,b_d_scale)
    convergence_index(1)=sum(abs(du(2:nt)))/abs(du(1))/(nt-1)
    write(0,*) '***ßsum|dut|/|du0|=',convergence_index(1)
    if (index_file.ne.'NULL' )call sep_write(convergence_index,index_file)

    write(0,*) 'sum(duu)=',sum(duu),'size(duu)',size(duu)
    duu=duu-sum(duu)/nt
    write(0,*) 'should be 0: sum(duu)=',sum(duu)
    !df=fnew-f=du*f+du*du*f/2=du*f*(1+du/2)


!     call FT1D(duu(:),du(:),nt,-1,b_d_scale)
!     du(1)=0.
!     call FT1D(du(:),duu(:),nt,1,f_d_scale)


    call D_FT1D(duu(:),du(:),nt,-1,b_d_scale)

!     do it=1,nt
! !       du(it)=sign(dc_one,du(it))
!         if (real(du(it)).gt.0.000001) then
!           du(it)=1.
!         else if (real(du(it)).lt.-0.000001) then
!           du(it)=-1.
!         else 
!           du(it)=0
!         end if
!     end do 

    du=du*mask_du

    if (n_sym_near_du.gt.0) then
     do it=2,n_sym_near_du
       dc_temp=(du(it)+du(nt-it+2))/2
       du(it)=dc_temp
       du(nt-it+2)=dc_temp
     end do
    end if

    write(0,*) 'max imag(du)=',maxval(abs(imag(du)))
    write(0,*) 'max real(du)=',maxval(abs(real(du)))

    if (du_file.ne.'NULL' )call sep_write(real(real(du)),du_file)

    !todo: time domain weighting on duu (gradient)

!     write(0,*) 'aaa',duu,'bbb'
    !$OMP parallel do private(ix)
    do ix=1,nx
      drr(:,ix)=rr(:,ix)*duu 
      call FT1D(drr(:,ix),dr(:,ix),nt,-1,b_d_scale)
    end do
    !$OMP end parallel do 

    dr=real(dr) !!!
     write(0,*) 'should be real:dr',dr(1:5,1)
    alpha=0.
!       temp1=0. !todo: check set 0 here
!       temp2=0.
    goto_updata_dr=.false.
    do iter_alpha=1,niter_alpha
        if (goto_updata_dr.eq..true.) then
          exit
        end if
        temp1=0. !todo: check set 0 here
        temp2=0.
      
      if (is_weight.eq..false.) then
        !$OMP parallel do private(ix,ithreads) REDUCTION(+:temp1,temp2)
        do ix=1,nx
          ithreads=omp_get_thread_num()+1
          call hybrid_d1(r(:,ix),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H'(r)
          call hybrid_d2(r(:,ix),wrk_2d_3(:,ithreads),threshold_traces(ix)) !wrk3=H''(r)
           !write(0,*) 'should be real:wrk2,wrk3',wrk2(1:5),wrk3(1:5),r(1:5,ix)
          temp1=-1.*sum((wrk_2d_2(:,ithreads)*dr(:,ix)))+temp1
          temp2=sum((wrk_2d_3(:,ithreads)*(dr(:,ix)*dr(:,ix))))+temp2
          !write(0,*) 'wrk3,dr,dr',r(:,ix)
        end do
        !$OMP end parallel do
!        write(0,*) 'r',r,'iter_alpha',iter_alpha
        inc_alpha=temp1/temp2
      else
        !$OMP parallel do private(ix,ithreads,weight_ix) REDUCTION(+:temp1,temp2)
        do ix=1,nx
          weight_ix=mod(ix-1,weight_n2)+1
          ithreads=omp_get_thread_num()+1
          wr(:,ix)=r(:,ix)*weight(:,weight_ix)
          call hybrid_d1(wr(:,ix),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H'(w*r)
          call hybrid_d2(wr(:,ix),wrk_2d_3(:,ithreads),threshold_traces(ix)) !wrk3=H''(w*r)
           !write(0,*) 'should be real:wrk2,wrk3',wrk2(1:5),wrk3(1:5),r(1:5,ix)
          temp1=temp1+1.*sum((wrk_2d_2(:,ithreads)*dr(:,ix)*weight(:,weight_ix)))
          temp2=temp2+sum((wrk_2d_3(:,ithreads)*(dr(:,ix)*dr(:,ix))*weight(:,weight_ix)*weight(:,weight_ix))) !w^2 for 2nd H''(w*r)
          !write(0,*) 'wrk3,dr,dr',r(:,ix)
        end do
        !$OMP end parallel do
!        write(0,*) 'r',r,'iter_alpha',iter_alpha
        inc_alpha=-temp1/temp2

      end if

      !re-calc the r, and see the improvement
      if (abs(prv_r_norm).gt.0) then
!         alpha_search=0.
        do while (.true.)
!           alpha_serch=alpha_search+1
          temp3=0.
!       write(0,*) 'debug a1'
          if (is_weight.eq..false.) then
            call calc_new_norm(inc_alpha,r,dr,wrk_2d_2,wrk_2d_3,temp3,threshold_traces)
          else
            call calc_new_norm(inc_alpha,r,dr,wrk_2d_2,wrk_2d_3,temp3,threshold_traces,weight)
          end if
!       write(0,*) 'debug a2'
!           do ix=1,nx
!             wrk3(:)=(r(:,ix)+inc_alpha*dr(:,ix))
!             call hybrid_d0(wrk3(:),wrk2(:)) !wrk2=H(r)
!             temp3=temp3+sum(wrk2(:))
!             !write(0,*) '       temp3',temp3,ix,inc_alpha,temp1,temp2!(r(:,ix)+inc_alpha*dr(:,ix))
!           end do
          ratio=(prv_r_norm-temp3)/prv_r_norm
!           write(0,*) 'ratio',abs(ratio),prv_r_norm,temp3
          if (real(ratio).gt.1e-8) then ! not over shot
            exit
          else if(real(ratio).gt.-1e-8 .or. abs(inc_alpha).lt.1e-8) then
            goto_updata_dr=.true. !!! converged go to updata dr
            write(0,*) '      converged go to updata dr'
            exit
          end if
          inc_alpha=inc_alpha/2
          write(0,*) '      over shot',inc_alpha
        end do

      end if



      prv_r=r
      if (abs(temp3).gt.0) then
        prv_r_norm=temp3
      else
          if (prv_r_norm.ne.0.) then
             call erexit( 'assert(prv_r_norm.ne.0.) failed')
!              stop
          end if

          if (is_weight.eq..false.) then
!       write(0,*) 'debug a4'
            call calc_new_norm(inc_alpha,r,dr,wrk_2d_2,wrk_2d_3,prv_r_norm,threshold_traces)
          else
!       write(0,*) 'debug a5'
            call calc_new_norm(inc_alpha,r,dr,wrk_2d_2,wrk_2d_3,prv_r_norm,threshold_traces,weight)
          end if
!           do ix=1,nx
!             wrk3(:)=r(:,ix)+inc_alpha*dr(:,ix)
!             call hybrid_d0(wrk3(:),wrk2(:)) !wrk2=H(r)
!             prv_r_norm=prv_r_norm+sum(wrk2(:))
!           end do
      end if
!       write(0,*) 'debug a3'

      alpha=alpha+inc_alpha !todo: less then first several iteration
      write(0,*) 'iter_alpha=',iter_alpha, alpha

      r=r+inc_alpha*dr
      !r=real(r)
    end do
!     call D_FT1D(duu(:),du(:),nt,-1,b_d_scale) !todo: this is not needed
!     write(0,*) duu
    u = u + alpha*du
    u = real(u) !todo

!     !!unifiy
!     u=u/u(1)

    

!     !!calc residual norm
!     do ix=1,nx
!       call hybrid_d0(r(:,ix),wrk(:,ix))
!     end do
!     res_norm(iter)=sum(abs(wrk))
!     write(0,*) 'r hybrid norm',res_norm(iter)

    if (linear_r_file.ne.'NULL' )call sep_write(real(r),linear_r_file)
    

    !! re-calc residual
    call D_FT1D(u(:),uu(:),nt,1,f_d_scale)
    write(0,*) 'Should be 0: sum(uu)=',sum(uu)
    euu=exp(uu)

   !$OMP parallel do private(ix)
    do ix=1,nx
      rr(:,ix)=dd(:,ix)*euu(:)
      call FT1D(rr(:,ix),r(:,ix),nt,-1,b_d_scale)
    end do
    !$OMP end parallel do
    res=real(r)!*input_file_scale
    write(0,*) 'write result'
    call sep_write(res)

    if (u_file.ne.'NULL' )call sep_write(real(real(u)),u_file)



    if (filter_file.ne.'NULL' ) then
      wrk2(:)=euu(:)
      call FT1D(wrk2(:),wrk3(:),nt,-1,b_d_scale)
!       wrk3=uu
      call sep_write(real(real(wrk3(:))),filter_file)
    end if

    if (l1_l2_file.ne.'NULL' ) then
      temp4=0.
      temp5=0.
      if (is_weight) then
        !$OMP parallel do private(ix,ithreads,weight_ix) REDUCTION(+:temp4,temp5)
        do ix=1,nx
          weight_ix=mod(ix-1,weight_n2)+1
          ithreads=omp_get_thread_num()+1
          wrk_2d_3(:,ithreads)=(r(:,ix)*r(:,ix))*weight(:,weight_ix)*weight(:,weight_ix)
          temp4=temp4+sum(wrk_2d_3(:,ithreads))
          temp5=temp5+sum(abs(r(:,ix))*weight(:,weight_ix))
        end do
        !$OMP end parallel do
      else
        !$OMP parallel do private(ix,ithreads) REDUCTION(+:temp4)
        do ix=1,nx
          ithreads=omp_get_thread_num()+1
          wrk_2d_3(:,ithreads)=r(:,ix)*r(:,ix)
          temp4=temp4+sum(wrk_2d_3(:,ithreads))
          temp5=temp5+sum(abs(r(:,ix)))
        end do
        !$OMP end parallel do
      end if
        
        temp4=temp5/sqrt(temp4)
        write(0,*) 'l1/l2'
      call sep_write(real(real(temp4)),l1_l2_file)
    end if


    write(0,*) 'end iter',iter,alpha
  end do

  if (res_norm_file.ne.'NULL' ) call sep_write(res_norm,res_norm_file)

!   do ix=1,nx
!     write(0,*) ix,mod(ix-1,weight_n2)+1,nx
!   end do

  call sep_close()

  deallocate(dd,uu,duu,drr,d,r,dr,u,du,rr,data,res,wrk2,wrk3,wrk4,res_norm,euu,prv_r,wrk_2d_2,wrk_2d_3,wrk_2d_4,wr,threshold_traces,mask_du,whwr)

  if (allocated(weight)) deallocate(weight)

end program

SUBROUTINE calc_new_norm(inc_alpha,r,dr,wrk_2d_2,wrk_2d_3,output,threshold_traces,weight)

  use hybrid
  implicit none
  optional :: weight
  double complex,intent(in)	:: inc_alpha
  double complex,intent(out)	:: output
  complex,dimension(:,:) 	:: r,dr
  complex,dimension(:,:) 	:: wrk_2d_2,wrk_2d_3
  real ,dimension(:,:),intent(in) :: weight
  real ,dimension(:),intent(in) ::threshold_traces
  integer	:: ix,ithreads

  integer       :: weight_n2,weight_ix

  integer, external	:: omp_get_thread_num,omp_get_num_threads
  
  output=0.

  
  if (present(weight)) then
    weight_n2=size(weight,2)
!     write(0,*) 'weight_n2',weight_n2
    !$OMP parallel do private(ix,ithreads,weight_ix) REDUCTION(+:output)
    do ix=1,size(r,2)
      weight_ix=mod(ix-1,weight_n2)+1
      ithreads=omp_get_thread_num()+1
      wrk_2d_3(:,ithreads)=(r(:,ix)+inc_alpha*dr(:,ix))*weight(:,weight_ix)
      call hybrid_d0(wrk_2d_3(:,ithreads),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H(wr)
      output=output+sum(wrk_2d_2(:,ithreads))
    end do
    !$OMP end parallel do
  else
    !$OMP parallel do private(ix,ithreads) REDUCTION(+:output)
    do ix=1,size(r,2)
      ithreads=omp_get_thread_num()+1
      wrk_2d_3(:,ithreads)=r(:,ix)+inc_alpha*dr(:,ix)
      call hybrid_d0(wrk_2d_3(:,ithreads),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H(r)
      output=output+sum(wrk_2d_2(:,ithreads))
    end do
    !$OMP end parallel do
  end if

end SUBROUTINE 

SUBROUTINE calc_current_norm(r,wrk_2d_2,wrk_2d_3,output,threshold_traces,weight)

  use hybrid
  implicit none
  optional :: weight
  double complex,intent(out)	:: output
  complex,dimension(:,:) 	:: r
  complex,dimension(:,:) 	:: wrk_2d_2,wrk_2d_3
  real ,dimension(:,:),intent(in) :: weight
  real ,dimension(:),intent(in) ::threshold_traces
  integer	:: ix,ithreads

  integer       :: weight_n2,weight_ix

  integer, external	:: omp_get_thread_num,omp_get_num_threads
  
  output=0.
  
  if (present(weight)) then
    weight_n2=size(weight,2)
!     write(0,*) 'weight_n2',weight_n2
    !$OMP parallel do private(ix,ithreads,weight_ix) REDUCTION(+:output)
    do ix=1,size(r,2)
      weight_ix=mod(ix-1,weight_n2)+1
      ithreads=omp_get_thread_num()+1
      wrk_2d_3(:,ithreads)=(r(:,ix))*weight(:,weight_ix)
      call hybrid_d0(wrk_2d_3(:,ithreads),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H(r)
      output=output+sum(wrk_2d_2(:,ithreads))
    end do
    !$OMP end parallel do
  else
    !$OMP parallel do private(ix,ithreads) REDUCTION(+:output)
    do ix=1,size(r,2)
      ithreads=omp_get_thread_num()+1
      wrk_2d_3(:,ithreads)=r(:,ix)
      call hybrid_d0(wrk_2d_3(:,ithreads),wrk_2d_2(:,ithreads),threshold_traces(ix)) !wrk2=H(r)
      output=output+sum(wrk_2d_2(:,ithreads))
    end do
    !$OMP end parallel do
  end if

end SUBROUTINE 

SUBROUTINE assert(expression,string)
  implicit none
  logical 	:: expression
  character (len=*) :: string
  if (expression.eq..false.) then
    call erexit(string)
  end if
end SUBROUTINE


! module plane_search_mod
! 
!   use hybrid
! 
!   implicit none
! 
!   real, dimension (:), allocatable, private   :: s, ss
!   double precision, dimension(:), allocatable,private     :: sds, gdg, gds
!   double precision, dimension(:), allocatable,private     :: c0,c1,c2
! 
! contains 
! 
! 
!   subroutine ps_init(len_mod,len_total_res)
!     if (allocated(ss)) then
!        deallocate(ss,s)
!        deallocate(error_cg,avg_rd,avg_rm)
!        deallocate(error_m, error_d)
!     end if
! 
!     allocate(ss(len_total_res),s(len_mod))
!     allocate (c0(len_total_res),c1(len_total_res),c2(len_total_res)) 
!     allocate (gdg(len_total_res),sds(len_total_res),gds(len_total_res)) 
!   end subroutine ps_init
! 
!   subroutine ps_close( )
!     if (allocated(ss)) then
!        deallocate(ss,s)
!        deallocate(gdg,sds,gds,c0,c1,c2)
!     end if
!   end subroutine ps_close
! 
!   integer function  solve_alfa_beta(forget,gg,alfa,beta)
!     complex,double precision,intent(out)   :: alfa, beta
!     complex,double precision               :: determ,c1dg,c1ds,c2dgg,c2dss,c2dgs
!     real :: sin0_2       !sin(theta)**2, indicate the orthoganlity of search directions
! 
!     real, dimension(:)  :: gg
!     logical, intent(in) :: forget
! 
!     solve_alfa_beta = 0
! 
!     ! === Pick Line Search or Plane Search
!     if( forget) then
!        !write(0,*) "enter line search, deepest descent"
!        ! Steepest Descent
!        beta = 0.0
! 
!        gdg   = dprod(gg,gg)
!        c2dgg = sum(dprod(c2,gdg))
! 
!        if( c2dgg == 0.0 )  then
! 	solve_alfa_beta = 1
!           call erexit('NORM_CGSTEP: Gradient vanishes identically: C"*gg*gg=0 .')
!        end if
! 
!        alfa = - sum( dprod(c1,gg)) / c2dgg
!     else
!        ! Plane Search for alpha & beta
!        !write(0,*) "enter the plane search"
!        c1dg = - sum(dprod(c1,gg))   ! c1dg = -C'*G
!        c1ds = - sum(dprod(c1,ss))   ! c1ds = -C'*S
! 
!        gdg = dprod(gg,gg)           ! search plane by solving 2-by-2
!        sds = dprod(ss,ss)           ! C"*G . (G*alfa + S*beta) = -C'*G
!        gds = dprod(gg,ss)           ! C"*S . (G*alfa + S*beta) = -C'*S
! 
!        c2dgg = sum(dprod(c2,gdg))   ! c2dgg = C"*G*G
!        c2dss = sum(dprod(c2,sds))   ! c2dss = C"*S*S
!        c2dgs = sum(dprod(c2,gds))   ! c2dgs = C"*G*S
! 
!        
! 
!        if ((c2dgg==0.d0) .OR. (c2dss==0.d0)) then
!           !Since C2 is always greater than zero, this case should not happen
!           write (0,*) 'Plane Search : Gradient vanishes identically: C"*gg*gg=0 OR C"*ss*ss=0., c2dgg,c2dss=',c2dgg,c2dss
!           solve_alfa_beta = 1
!           !call erexit('Plane Search : Gradient vanishes identically')
!           return
!        end if
! 
!        !write(0,*) "bivariate search, sum(gds), sum(gdg), sum(sds) =",sum(gds),sum(gdg),sum(sds)
! 
!        ! Here compute the extent of orthogonality
!        ! cos(theta) = (AdotB)/(|A||B|)
!        sin0_2 = 1- (sum(gds)/sum(gdg))*(sum(gds)/sum(sds))
!        if (sin0_2 < 1e-3) then
!           write(0,*) "g and s are highly linear dependent! plane search ceased!"
!           solve_alfa_beta = 1
!        end if
!        
!        
!        determ = 1.d0 - (c2dgs/c2dgg)*(c2dgs/c2dss)
!        !write (0,*) "the determ scale is=", determ, "sin0^2=", sin0_2
! 
!        determ = c2dgg * c2dss * determ
! 
!        if (determ < 1.d-12) then
!           write(0,*) "c2dgg, c2dss, sin^2=", c2dgg, c2dss, determ
!           write(0,*) "the determinant of C_2 matrix is too small! plane search ceased"
!           solve_alfa_beta = 1
!        end if
! 
!        alfa = ( c2dss * c1dg - c2dgs * c1ds ) / determ
!        beta = (-c2dgs * c1dg + c2dgg * c1ds ) / determ
! 
!     end if
!   end function solve_alfa_beta
! end module plane_search_mod
