!!$
!!$=head1 NAME
!!$
!!$ shift.x - circular shift the dataset fastest dimension.
!!$   Written by Qiang Fu
!!$   qiang@sep.stanford.edu 
!!$   Version 1.0
!!$   June, 2011
!!$
!!$=head1 DESCRIPTION
!!$   Before
!!$   [d0] [d1] [d2] [d3] ... [dn]
!!$    |    |    |    |   ...  |
!!$    t0   t1   t2   t3  ...  tn
!!$
!!$   After shift=1
!!$   [d1] [d2] [d3] [d4] ... [d0]
!!$    |    |    |    |   ...  |
!!$    t0   t1   t2   t3  ...  tn
!!$
!!$=head2 SYNOPSIS 
!!$
!!$ Shift the dataset along negtive direction: 
!!$   shift.x <in.H  shift=0 change_axis=(0,[1]) keep_first=(0,[1]) > out.H 
!!$
!!$=head3 PARAMETERS FOR FORWARD TOMOGRAPHY
!!$
!!$ shif- integer
!!$      The time simples to be shifted toward negtive direction
!!$
!!$ change_axis- integer
!!$      [1] change the axis (the o1) accordingly
!!$      0   do not change axis
!!$
!!$ keep_first- integer
!!$      [1] change the axis (the o1) accordingly
!!$      0   do not change axis!!$
!!$
!!$
!!$
!!$=head1 COMMENTS
!!$
!!$ 
!!$
!!$=cut

program shift_x
  use sep
  use util_axisIO

  implicit none


  integer :: naxis,n1,n2,i,j
  type(myaxis)  axis1, axis2, axis3
  ! <in.H bubble > out.H pad=pad_times
  ! pad_times means to pad as n times as input length 0 after input data BEFORE inverse.
  !           That will affect the negtive time casued the Periodicity



 

  real, dimension(:,:), allocatable :: d,output
  integer :: shift
  
   character(len=256) :: csz_shift, csz_change_axis, csz_keep_axis_first_value

  integer change_axis, keep_axis_first_value

  csz_shift = "shift"
  csz_change_axis="change_axis"
  csz_keep_axis_first_value="keep_first"


   !Params
  call sep_init(SOURCE)


  call from_param (csz_shift, shift, 0)

  call from_param (csz_change_axis, change_axis, 1)
  call from_param (csz_keep_axis_first_value, keep_axis_first_value, 1)

  call hinaxis(axis1,1)
  call hinaxis(axis2,2)
  call hinaxis(axis3,3)
  n1=axis1%n
  n2=axis2%n
  if (axis3%n.gt.1) then
    n2=n2*axis3%n
  end if


!   write(0,*) " <in.H shift shift=0 change_axis=[0,(1)] keep_first=[0,(1)] > out.H "
  ! pad_times means to pad as n times as input length 0 after input data BEFORE inverse.
  !           That will affect the negtive time casued the Periodicity








  if (n1.le.1 .and. n2.le.1) then
    write(0,*) "n1 and n2 <= 0"
    stop
  end if

  allocate (d(n1,n2))
  
  allocate (output(n1,n2))
 

  call sep_read(d)

  if (shift.eq.0) then
    shift=n1/2
  end if

  if (abs(shift).ge.n1) then
    write(0,*) "abs(shift).ge.n1",abs(shift),n1
    stop
  end if
  !

  if (change_axis.ne.0) then
    axis1%o=axis1%o+axis1%d*shift
!     write(0,*) axis1%o ,axis1%o ,n1 ,axis1%d*n1/2
    if (keep_axis_first_value.ne.0 .and. shift .gt. 0) then
      axis1%o=axis1%o-axis1%d*axis1%n
    end if
    call houtaxis(axis1,1)
  end if



  !flip
  if (shift.ge.1) then
    output(n1-shift+1:n1,:)=d(1:shift,:)
    output(1:n1-shift,:)=d(shift+1:n1,:)
  else if (shift.le.-1) then
    output(1:-shift,:)=d(n1+shift+1:n1,:)
    output(-shift+1:n1,:)=d(1:n1+shift,:)
  else
    write(0,*) "shift.eq.0",shift
    stop
  end if


 

  call sep_write(output)
  deallocate(d,output)

  call sep_close()
end program
